; Serial test program: Send 'A' with 9600 8N1 once, then repeat
; receiving a character and echoing it.
;
; This program can be run from ROM or RAM.

; /usr/frankenstein/bin/as8048 -o echo.hex -l echo.list echo.asm
;
; P1.0: enable multiplexer, active low
; P1.1: A17 for code access
; P1.2: A17 for data access
; P1.3: A16 for data access
; P1.4-P1.7: A12-A15 for code access
;
; P2.0-P2.7: A8-A15 for data access

		org 0

		ent0 clk		; Crystal/3 on T0
		anl p1,#3		; a16data=0 a17data=0 a17code=* mux=*
;
; UART setup
;
		mov a,#$80		; A15-A8 for data access (UART)
		outl p2,a

		mov a,#$00		; Turn off interrupts
		mov r0,#$01		; in UART register 1
		movx @r0,a

		mov a,#$80		; set DLAB on
		mov r0,#$03		; in UART register 3
		movx @r0,a

		mov a,#13		; baud rate divisor lsb (9600)
		mov r0,#$00		; in UART register 0
		movx @r0,a
		mov a,#$00		; baud rate divisor msb (9600)
		mov r0,#$01		; in UART register 1
		movx @r0,a
	
		mov a,#$03		; 8N1
		mov r0,#$03		; in UART register 3
		movx @r0,a

		mov a,#$c7		; FIFO control
		mov r0,#$02		; in UART register 2
		movx @r0,a

		mov a,#$0b		; set DTR, RTS, and OUT2
		mov r0,#$04		; in UART register 4
		movx @r0,a
;
; send byte
;
		mov r0,#$00		; UART register 0
		mov a,#65		; 'A'
		movx @r0,a		; write
;
; receive byte and send it back
;
receive
		mov r0,#$05		; in UART register 4
		movx a,@r0
		inc a			; toggle bit 0
		jb0 receive		; toggled bit high -> original bit low
		mov r0,#$00		; from UART register 0
		movx a,@r0		; read character
		movx @r0,a		; write back

		jmp receive
