; I8HEX/INTEL 8 loader

; /usr/frankenstein/bin/as8048 -o load.hex -l load.list load.asm

; init
;
        org     0
        jmp	go
booted
        mov     a,#2            ; led_a16data, a17data, A17CODE, multiplexer
	outl	p1,a
go
	ent0	clk		; Crystal/3 on T0
        mov     a,#0            ; led_a16data, a17data, a17code, multiplexer
        outl    p1,a
        jmp     main

; db 65

; read char, return in A
; destroy R0
; 19 cycles self

getchar
	anl	p1,#3		; clear led_a16data, a17data, keep a17code, multiplexer
	mov	a,#$80		; A15-A8 for data access
	outl	p2,a
	mov	r0,#5		; Line Status Register
receive
	movx	a,@r0
	inc	a		; toggle bit 0
	jb0	receive		; toggled bit high -> original bit low
	mov	r0,#$00		; from UART register 0
	movx	a,@r0		; read character
	ret

; put char
; restroy R0, R1

putchar
	xch	a,r1
	anl	p1,#3		; clear led_a16data, a17data, keep a17code, multiplexer
	mov	a,#$80		; A15-A8 for data access
	outl	p2,a
        mov	r0,#5		; Line Status Register
full
	movx	a,@r0
	anl	a,#$20		; Empty Transmitter Holding Register?
	jz	full
	xch	a,r1
	mov	r0,#$00		; UART register 0
	movx	@r0,a		; write
	ret

; read hex letter, return decoded nibble in A
; destroy R0
; at worst 21 cycles self + 19 cycles getchar = 40 cycles

getnibble
	call	getchar
	add	a,#208		; map '0' to 0
	jnc	nohex		; no carry means less than '0'
	add	a,#246		; map 0 to 246 and 9 to 255
	jc	alpha		; carry means greater than 9
	add	a,#10		; map 0 to 0, 9 to 9
	ret
alpha
	add	a,#249		; 'A' is 7, map 'A' to 0
	anl	a,#$df		; map lower to upper case
	add	a,#250		; map 'F' to 255
	jc	nohex		; carry means greater than 'F'
	add	a,#16		; map 'F' to 15
	ret
nohex
	ret

; read hex pair, update checksum, return decoded byte in A
; destroy R0,R1
; 12 cycles self + 2*37 cycles getnibble = 86 cycles

getbyte
	call	getnibble
	swap	a
	mov	r1,a
	call	getnibble
	orl	a,r1		; decoded value in A
	xch	a,r2		; exchange value and checksum
	add	a,r2		; add value
	xch	a,r2		; exchange value and checksum again
	ret

; load file
; destroys r0,r1,r2,r3

main

;
; UART setup
;
	anl	p1,#3		; clear led_a16data, a17data, keep a17code, multiplexer
	mov	a,#$80		; A15-A8 for data access
	outl	p2,a

	mov	a,#$00		; Turn off interrupts
	mov	r0,#$01		; in UART register 1
	movx	@r0,a

	mov	a,#$80		; set DLAB on
	mov	r0,#$03		; in UART register 3
	movx	@r0,a

	mov	a,#13		; baud rate divisor lsb (9600)
	mov	r0,#$00		; in UART register 0
	movx	@r0,a
	mov	a,#$00		; baud rate divisor msb (9600)
	mov	r0,#$01		; in UART register 1
	movx	@r0,a
	
	mov	a,#$03		; 8N1
	mov	r0,#$03		; in UART register 3
	movx	@r0,a

	mov	a,#$c7		; FIFO control
	mov	r0,#$02		; in UART register 2
	movx	@r0,a

	mov	a,#$0b		; set DTR, RTS, and OUT2
	mov	r0,#$04		; in UART register 4
	movx	@r0,a


	mov	a,#35
	call	putchar
	mov	r2,#0		; checksum = 0
getline
	call	getchar
	add	a,#198		; sub ':'
	jnz	nocolon
	call	getbyte		; get count in r3
	mov	r3,a
	call	getbyte		; get offset high byte in r4
	mov	r4,a
	call	getbyte		; get offset low byte in r5
	mov	r5,a
	call	getbyte		; get record type
	jz	datarec		; type 0 (data)
        dec     a
	jz	endrec		; type 1 (end)
	jmp	norec
datarec
	call	getbyte
	mov	r1,a		; save data byte in r1
	orl	p1,#4		; keep a16data, set A17DATA, keep a17code, mux
        anl     p1,#7           ; clear a16data, keep A17DATA, a17code, mux
	mov	a,r4		; offset high byte
	outl	p2,a
	mov	a,r5		; offset low byte
	mov	r0,a
	mov	a,r1		; store data byte
	movx	@r0,a
	mov	a,r5		; increment offset
	add	a,#1		; inc does not set carry, add does
	mov	r5,a
	mov	a,r4
	addc	a,#0
	mov	r4,a
	djnz	r3,datarec
	call	checksum
	jmp	getline
endrec
	call	checksum

 mov a,#122
 call putchar

	jmp	booted

checksum
	call	getbyte
	ret

norec
 mov a,#48
 call putchar
	jmp	norec
nocolon
	jmp	getline
