; External RAM tester, may be run from ROM or RAM.
;
; Fill 64k of XRAM with LSB XOR MSB of its address, check if it is filled,
; output each failed address and finally output OK.  The test takes only
; a few seconds.
; If you run it from ROM, start may be set to 0, otherwise page 0 containing
; the code will not be tested.
;
; /usr/frankenstein/bin/as8048 -o xram.hex -l xram.list xram.asm


start	equ	1
stop	equ	256

; init
;
        org     0
	ent0	clk		; Crystal/3 on T0
        jmp     main

; put char
; restroy R0, R1
;
putchar
	xch	a,r1
	anl	p1,#3		; a16data=0 a17data=0 a17code=* mux=*
	mov	a,#$80		; A15-A8 for data access
	outl	p2,a
        mov	r0,#5		; Line Status Register
full
	movx	a,@r0
	anl	a,#$20		; Empty Transmitter Holding Register?
	jz	full
	xch	a,r1
	mov	r0,#$00		; UART register 0
	movx	@r0,a		; write
	ret

; output byte as hex
; destroys r0,r1,r2
;
puthex
		mov r2,a
		anl a,#$f0
		swap a
		add a,#246
		jnc digith
		add a,#7	; 'A' - '0'
digith		add a,#58	; '0' + 10
		call putchar
		mov a,r2
		anl a,#$0f
		add a,#246
		jnc digitl
		add a,#7	; 'A' - '0'
digitl		add a,#58	; '0' + 10
		call putchar
		ret

main

;
; UART setup
;
	anl	p1,#3		; a16data=0 a17data=0 a17code=* mux=*
	mov	a,#$80		; A15-A8 for data access
	outl	p2,a

	mov	a,#$00		; Turn off interrupts
	mov	r0,#$01		; in UART register 1
	movx	@r0,a

	mov	a,#$80		; set DLAB on
	mov	r0,#$03		; in UART register 3
	movx	@r0,a

	mov	a,#13		; baud rate divisor lsb (9600)
	mov	r0,#$00		; in UART register 0
	movx	@r0,a
	mov	a,#$00		; baud rate divisor msb (9600)
	mov	r0,#$01		; in UART register 1
	movx	@r0,a
	
	mov	a,#$03		; 8N1
	mov	r0,#$03		; in UART register 3
	movx	@r0,a

	mov	a,#$c7		; FIFO control
	mov	r0,#$02		; in UART register 2
	movx	@r0,a

	mov	a,#$0b		; set DTR, RTS, and OUT2
	mov	r0,#$04		; in UART register 4
	movx	@r0,a

;
; Fill external RAM.
;
	mov a,#start
	mov r6,a
	mov a,#0
	mov r7,a
	anl p1,#3		; a16data=0 a17data=1 a17code=* mux=*
	orl p1,#4
	mov a,r6
	outl p2,a
test
	mov a,r7
	mov r0,a
	xrl a,r6
	movx @r0,a
	mov a,r7
	add a,#1
	mov r7,a
	mov a,r6
        jnc samepage
	inc a
	mov r6,a
	outl p2,a
samepage
	add a,#256-stop
	jnz test

; verify
	mov a,#start
	mov r6,a
	mov a,#0
	mov r7,a
	anl p1,#3		; a16data=0 a17data=1 a17code=* mux=*
	orl p1,#4
check
	mov a,r6
	outl p2,a
	mov a,r7
	mov r0,a
	movx a,@r0
	xrl a,r6
	xrl a,r7
	jnz fail
cont
	mov a,r7
	add a,#1
	mov r7,a
	mov a,r6
	jnc samepage2
	inc a
	mov r6,a
	outl p2,a
samepage2
	add a,#256-stop
	jnz check

	mov a,#'O'
	call putchar
	mov a,#'K'
	call putchar
halt
	jmp halt

fail
	mov a,#78
	call putchar
	anl p1,#3		; a16data=0 a17data=1 a17code=* mux=*
	orl p1,#4
	jmp cont


