; /usr/frankenstein/bin/as8048 -o f.hex -l f.list f.asm

; System vectors and boot code ;{{{Lisp}}};{{{
;
; This is the entry point after a hardware reset or a software code
; segment switch.  After a hardware reset, all bits of P1 are set high.
; That disables the address multiplexer and executes code from ROM.
; The early boot code checks the P1 bit for the address multiplexer to
; see if it runs from ROM or RAM, allowing to use the same code for both.

	        org 0
	        jmp reset

		org 3
		jmp interrupt

		org 7			; timer interrupt handler
		sel rb1
		mov r0,#CLOCKTICKS
		inc @r0
		sel rb0
		retr

reset
		in a,p1
		jb0 hardwarereset	; A hardware reset sets all bits high
		jmp forth
;
; A hardware reset brought us here.  Configure T0 to provide a system
; clock signal for the UART and for peripheral cards.
;
hardwarereset
	        ent0 clk		; Crystal/3 on T0
	        clr a			; a16data=0 a17data=0 a17code=0 mux=0
	        outl p1,a		; (enable mux)
;
; Initialise UART to 9600 8N1.
;
		anl p1,#3		; a16data=0 a17data=0 a17code=* mux=*
		mov a,#$80		; A15-A8 for data access
		outl p2,a

		mov a,#$00		; Turn off interrupts
		mov r0,#$01		; in UART register 1
		movx @r0,a

		mov a,#$80		; set DLAB on
		mov r0,#$03		; in UART register 3
		movx @r0,a

		mov a,#13		; baud rate divisor lsb (9600)
		mov r0,#$00		; in UART register 0
		movx @r0,a
		mov a,#$00		; baud rate divisor msb (9600)
		mov r0,#$01		; in UART register 1
		movx @r0,a
	
		mov a,#$c7		; FIFO control
		inc r0			; in UART register 2
		movx @r0,a

		mov a,#$03		; 8N1
		inc r0			; in UART register 3
		movx @r0,a

		mov a,#$0b		; set DTR, RTS, and OUT2
		inc r0			; in UART register 4
		movx @r0,a

	        mov r0,#2		; FIFO Control Register
 		mov a,#3		; Enable FIFO and clear receive FIFO
		movx @r0,a
;
; This is supposed to suppress receiving a junk character after reset,
; but does not work.  Myth or recommended, it won't do any harm.
;
 		mov a,#5		; read LSR
		movx a,@r0

 		mov a,#6		; read MSR
		movx a,@r0
;
; This helps, though.
;
	        mov a,#1
		call timedgetchar
;
; Output the board name to show the SBC is alive.
;
		mov a,#'S'
		call putchar
		mov a,#'B'
		call putchar
		mov a,#'C'
		call putchar
		mov a,#'M'
		call putchar
		mov a,#'C'
		call putchar
		mov a,#'S'
		call putchar
		mov a,#'4'
		call putchar
		mov a,#'8'
		call putchar
;
; Wait a few seconds to receive a character.  'b' loads an Intel hexfile
; into RAM.  'r' loads the ROM into RAM.  After doing so, after receiving any
; other character or after a timeout, program control is passed to RAM.
; That way you can load whatever you want into RAM and the SBC will start
; it after being switched on, but you can always go back to the ROM.
;
		mov a,#10
		call timedgetchar
		mov r2,a

		mov a,#13
		call putchar
		mov a,#10
		call putchar

		mov a,r2
		xrl a,#'b'			; b (boot) loads Intel hexfile
		jz goi8hexload
		mov a,r2
		xrl a,#'r'			; r (reset) loads ROM
		jz romload
		mov a,#6			; any other starts from RAM
		jmp codesegswitch
goi8hexload	jmp i8hexload
;}}}
romload ;{{{ load ROM to RAM
		clr a
		mov r0,a		; low byte
		mov r1,a		; high byte
		outl p2,a
		mov r2,#1+ HIGH forthend; counter in pages
copyloop	anl p1,#0		; a16data=0 a17data=0 a17code=0 mux=0
		movx a,@r0
		orl p1,#4		; a16data=* a17data=1 a17code=* mux=*
		movx @r0,a
		djnz r0,copyloop
		inc r1
		mov a,r1
		outl p2,a
		djnz r2,copyloop
		mov a,#6
		jmp codesegswitch
;}}}
i8hexload ;{{{ load I8HEX/INTEL 8 file to RAM
		dis i
		dis tcnti
		stop tcnt
		anl p1,#0		; a16data=0 a17data=0 a17code=0 mux=0

		mov r2,#0		; checksum = 0
getline
		call getchar
		add a,#198		; sub ':'
		jnz nocolon
		call getbyte		; get count in r3
		mov r3,a
		call getbyte		; get offset high byte in r4
		mov r4,a
		call getbyte		; get offset low byte in r5
		mov r5,a
		call getbyte		; get record type
		jz datarec		; type 0 (data)
	        dec a
		jz endrec		; type 1 (end)
		jmp norec
datarec
		call getbyte
		mov r1,a		; save data byte in r1
		orl p1,#4		; a16data=* a17data=1 a17code=* mux=*
	        anl p1,#7		; a16data=0 a17data=* a17code=* mux=*
		mov a,r4		; offset high byte
		outl p2,a
		mov a,r5		; offset low byte
		mov r0,a
		mov a,r1		; store data byte
		movx @r0,a
		inc r5
		mov a,r5
		jnz r4unchanged
		inc r4
r4unchanged	djnz r3,datarec
		call checksum
		jmp getline
endrec
		call checksum

		mov a,#6		; a16data=0 a17data=0 a17code=1 mux=0
		jmp codesegswitch

checksum
		call getbyte
		call getchar
		ret

norec
		mov a,#48
 		call putchar
		jmp norec
nocolon
		jmp getline

; read hex letter, return decoded nibble in A
; destroy R0
; at worst 21 cycles self + 19 cycles getchar = 40 cycles

getnibble
		call getchar
		add a,#208		; map '0' to 0
		jnc nohex		; no carry means less than '0'
		add a,#246		; map 0 to 246 and 9 to 255
		jc alpha		; carry means greater than 9
		add a,#10		; map 0 to 0, 9 to 9
		ret
alpha
		add a,#249		; 'A' is 7, map 'A' to 0
		anl a,#$df		; map lower to upper case
		add a,#250		; map 'F' to 255
		jc nohex		; carry means greater than 'F'
		add a,#16		; map 'F' to 15
nohex		ret

; read hex pair, update checksum, return decoded byte in A
; destroy R0,R1
; 12 cycles self + 2*37 cycles getnibble = 86 cycles

getbyte
		call getnibble
		swap a
		mov r1,a
		call getnibble
		orl a,r1		; decoded value in A
		xch a,r2		; exchange value and checksum
		add a,r2		; add value
		xch a,r2		; exchange value and checksum again
		ret
;}}}
putchar ;{{{ print character in a
; destroy R0, R1

		xch a,r1
		anl p1,#3	; a16data=0 a17data=0 a17code=* mux=*
		mov a,#$80	; A15-A8 for data access
		outl p2,a
	        mov r0,#5	; Line Status Register
full
		movx a,@r0
		anl a,#$20	; Empty Transmitter Holding Register?
		jz full
		xch a,r1
		mov r0,#0	; UART register 0
		movx @r0,a	; write
	        orl p1,#4	; a16data=0 a17data=1 a17code=* multiplexer=*
		ret
;}}}
getchar ;{{{ get character in a (17)
; destroys r0

		anl p1,#3	; a16data=0 a17data=0 a17code=* mux=*
		mov a,#$80	; A15-A8 for data access
		outl p2,a
getloop
		mov r0,#5	; Line Status Register (LSR #5)
		movx a,@r0
		dec r0          ; Modem Status Register (MCR #4)
		jb0 gotchar

		movx a,@r0	; set RTS (MCR #4.1)
		orl a,#2
		movx @r0,a

		jmp getloop

gotchar		
		movx a,@r0	; clear RTS (MCR #4.1)
		anl a,#253
		movx @r0,a

		mov r0,#0	; from UART register 0
		movx a,@r0	; read character
	        orl p1,#4	; a16data=0 a17data=1 a17code=* mux=*
		ret
;}}}
timedgetchar ;{{{ get character in a, timeout in a
; destroy r0,r2,r
		mov r2,a
		anl p1,#3	; a16data=0 a17data=0 a17code=* mux=*
		mov a,#$80	; A15-A8 for data access
		outl p2,a
receiveagain3
		mov r3,#128
receiveagain2
		mov r4,#0
receiveagain
		mov r0,#4	; set RTS in Modem Control Register (MCR #4.1)
		mov a,#2
		movx @r0,a

		inc r0		; move to Line Status Register (LSR #5)

		movx a,@r0
		jb0 gotone	; toggled bit high -> original bit low
		clr a
		djnz r4,receiveagain
		djnz r3,receiveagain2
		djnz r2,receiveagain3
		jmp gotauto
gotone
		dec r0		; back to Modem Control Register (MCR #4)
		clr a		; clear all bits, in particular RTS (MCR #4.1)
		movx @r0,a

		mov r0,#0	; from UART register 0
		movx a,@r0	; read character
gotauto
	        orl p1,#4	; a16data=0 a17data=1 a17code=* mux=*
		ret
;}}}

; The threading model ;{{{
;
; For primitive words, the code field stores machine code, for Forth
; words, it stores Forth code.  Forth code is directly threaded and the
; virtual Forth machine recognizes primitive words by having a value
; within the first two kilobytes.  The MCS-48 needs internal bank switching
; to address more than 2KB program space, and external bank switching to
; address more than 4KB.  By keeping the core small, the virtual Forth
; machine does not need any bank switching.
;
; To save a cell per word, this implementation does not use an interpreting
; function.  Instead, variables require a new primitive that pushes a pointer
; behind the created word to the data stack.  DOES> words do the same and need
; a branch to the code.
;}}}

; The Forth virtual machine
;
; This is also called the inner loop.  It consists of two entry points:
;
; "next" performs the fetch cycle of the virtual machine:
; It gets the next word in r2 r3 and increments the IP in r4 r5, followed
; by the execute cycle.  Register pairs are written in little endian order.
;
; "executeword" decides if the word is a primitive or word written in
; Forth by the value of the word.  If it is inside the first four KByte,
; it is a primitive.  Later words must be written in Forth.  That forbids
; to add words in assembler later, but it saves external bank switching
; and allows flat addressing of external memory.
;
; The MCS-48 does not contain an indirect long jump instruction, only
; a double indirect short jump.  For that reason, the next instruction
; is called to increase the stack pointer, the stack entry is overwritten
; by the primitive, and a return transfers control to it.  When the word
; is done, it jumps back to one of the three entry points above and the
; operation is complete.
;
; Forth words are run by pushing the IP in r4 r5 on the return stack
; and loading the IP with the new word, completing the operation by a
; jump to "next".  Forth words return using "EXIT", which pops
; r4 r5 off the return stack.

; 39/47 cycles for primitives, 48/58 cycles for Forth words

; The main pointers ;{{{
;
; The MCS-48 contains a small hardware stack that is used for
; subroutines of the assembler core.  It is not suitable for one
; of the Forth stacks.  They are kept in the internal RAM instead:

SP		equ 32			; grows up
RP		equ 34			; grows down
IP		equ 36
CLOCKTICKS	equ 38
IRAMLEN		equ 40

; All stack pointers point at the last element pushed for being
; able to fetch it without pointer arithmetic.
;
; The data area pointer DP is stored in external RAM as variable.
;}}}

forth ;{{{ system entry, falls through to next
;
; Now determine the amount of IRAM.  Small RAMs will ignore the high address
; bits and modify R1, whereas larger RAMs would modify the given address.
;
		mov r1,#255		; Store 255 in R1
		mov r0,#129		; Store 127 in R1+128
		mov a,#$7f
		mov @r0,a
		mov r0,#65		; Store 63 in R1+64
		mov a,#$3f
		mov @r0,a
		mov a,r1
		mov r0,#IRAMLEN
		mov @r0,a
;
; Initialise SP0, RP0, IP and go!
;
		mov r3,#6
		mov r2,#HIGH var_SP0
		mov r1,#LOW var_SP0
		mov r0,#SP
init_loop	mov a,r2
		outl p2,a
		movx a,@r1
		mov @r0,a
		inc r0
		inc r1
		mov a,r1
		jnz init_keepr2
		inc r2
init_keepr2	djnz r3,init_loop
		en tcnti
		strt t

		mov r0,#IP
		mov a,@r0
		mov r4,a
		inc r0
		mov a,@r0
		mov r5,a
;}}}
next ;{{{ r2 r3 = word at r4 r5, increase r4 r5 (17/18)
		mov a,r5		; (1) r2 r3 = r4 r5 @
		outl p2,a		; (2)
		mov a,r4		; (1)
		mov r1,a		; (1)
		movx a,@r1		; (2)
		mov r2,a		; (1)
		inc r1			; (1) increment IPL
		movx a,@r1		; (2)
		mov r3,a		; (1)
		inc r1			; (1) increment IPL
		mov a,r1		; (1)
		mov r4,a		; (1)
		jnz ipinc_done		; (2)
		inc r5			; (1)
ipinc_done

; FALLTHROUGH
;}}}
executeword ;{{{ execute word in r2 r3 (16 primitive, 31/33 Forth)
		mov a,r3	; (1)
		anl a,#$f8	; (2) mask out last three bits = 2 KB primitives
		jz primitive	; (2)

		mov r0,#RP	; (2) decrease RP
		mov a,@r0	; (1)
		add a,#254	; (2)
		mov r1,a	; (1) r1 = RP low byte
		mov @r0,a	; (1)
		inc r0		; (1)
		mov a,@r0	; (1)
		jc exec_norphdec; (2)
		dec a		; (1)
		mov @r0,a	; (1)
exec_norphdec
		outl p2,a	; (2) store IP where RP points to
		mov a,r4	; (1)
		movx @r1,a	; (2)
		inc r1		; (1)
		mov a,r5	; (1)
		movx @r1,a	; (2)
		mov a,r2	; (1) IP = word
		mov r4,a	; (1)
		mov a,r3	; (1)
		mov r5,a	; (1)
		jmp next	; (2)

primitive
		call foo	; (2) increase stack pointer
foo		mov r0,#8	; (2) r0 points to lowest stack element
		mov a,r2	; (1) *r0 = r2
		mov @r0,a	; (1)
		inc r0		; (1) ++r0
		mov a,r3	; (1) *r0 = r3
		mov @r0,a	; (1)
		ret		; (2) jump to primitive, which jumps back to next
;}}}
	reserve 8
ucmp16 ;{{{ unsigned comparison of r2 r3 and r4 r5, a=0 for equal, c for less than
; adding the two's complement gives the same result, but does not
; propagate carry if the subtrahend overflows to the addition.

		mov a,r3
		cpl a
		add a,r5
		cpl a
		jz comparelow
		ret
comparelow	mov a,r2
		cpl a
		add a,r4
		cpl a
		ret
;}}}
scmp16 ;{{{ signed comparison of r2 r3 and r4 r5, a=0 for equal, c for less than
; for different signs, sign decides result, for same signs, compare unsigned
		mov a,r5
		jb7 cmpx1
		mov a,r3		; x0
		jb7 cmp10
		jmp ucmp16		; 00, perform unsigned comparison
cmp10		clr c			; 10, set C, A is still NZ
		cpl c
		ret
cmpx1		mov a,r3
		jb7 ucmp16		; 11, perform unsigned comparison
		clr c			; 01, clear C and set A NZ
		mov a,r5
		ret
;}}}
rpop ;{{{ pop r2 r3 from return stack (21/23)
		mov r0,#RP+1	; (1)
		mov a,@r0	; (1)
		outl p2,a	; (2)
		dec r0		; r0=&IP low byte (1)
		mov a,@r0	; r1=*r0 (1)
		mov r1,a	; (1)
		movx a,@r1	; (2)
		mov r2,a	; (1)
		inc r1		; (1)
		movx a,@r1	; (2)
		mov r3,a	; (1)
		inc @r0		; (1)
		inc @r0		; (1)
		mov a,@r0	; (1)
		jnz rpop_done	; (2)
		inc r0		; (1)
		inc @r0		; (1)
rpop_done	ret		; (2)
;}}}
rpush ;{{{ push r2 r3 on return stack (22/24)
		mov r0,#RP	; (2)
		mov a,@r0	; (1)
		add a,#254	; (2)
		mov r1,a	; (1) r1 = RP low byte
		mov @r0,a	; (1)
		inc r0		; (1)
		mov a,@r0	; (1)
		jc rpdec_done	; (2)
		dec a		; (1)
		mov @r0,a	; (1)
rpdec_done	outl p2,a	; (2)
		mov a,r2	; (1)
		movx @r1,a	; (2)
		inc r1		; (1)
		mov a,r3	; (1)
		movx @r1,a	; (2)
		ret		; (2)
;}}}
fetchret ;{{{ r2 r3 = *(*r0) (15), fall through to dpush
; called with R0 pointing to high byte, returning with R0 pointing to low byte
		mov r0,#RP+1	; (1)
		mov a,@r0	; (1)
		outl p2,a	; (2)
		dec r0		; r0=&IP low byte (1)
		mov a,@r0	; r0=*r0 (1)
		mov r1,a	; (1)
		movx a,@r1	; (2)
		mov r2,a	; (1)
		inc r1		; (1)
		movx a,@r1	; (2)
		mov r3,a	; (1)
;}}}
dpush ;{{{ push r2 r3 on data stack (22/24)
		mov r0,#SP	; (2)
		inc @r0		; (1)
		inc @r0		; (1)
		mov a,@r0	; (1)
		inc r0		; (1)
		jnz dpinc_done	; (2)
		inc @r0		; (1)
dpinc_done	mov a,@r0	; (1)
		outl p2,a	; (2)
		dec r0		; (1)
		mov a,@r0	; (1)
		mov r0,a	; (1)
		mov a,r2	; (1)
		movx @r0,a	; (2)
		inc r0		; (1)
		mov a,r3	; (1)
		movx @r0,a	; (2)
		ret		; (2)
;}}}
dpopfetch ;{{{ pop cell from data stack in r4 5, fall through to dfetch (47/51)
	call saveIP
	call dpop	; (2+22/26)
	mov a,r2	; (1)
	mov r4,a	; (1)
	mov a,r3	; (1)
	mov r5,a	; (1)
; FALLTHROUGH
;}}}
dfetch ;{{{ fetch top of data stack and store in r2 r3, addr in r6 r7 (19)
		mov r0,#SP+1	; (2)
		mov a,@r0	; (1)
		mov r7,a	; (1)
		outl p2,a	; (2)
		dec r0		; (1)
		mov a,@r0	; (1)
		mov r6,a	; (1)
		mov r0,a	; (1)
		movx a,@r0	; (2)
		mov r2,a	; (1)
		inc r0		; (1)
		movx a,@r0	; (2)
		mov r3,a	; (1)
		ret		; (2)
;}}}
dpop2 ;{{{ pop cell from data stack and store in r4 r5, fall through to dpop (42/50)
		call saveIP
		mov r0,#SP+1		; (2) (20/24)
		mov a,@r0		; (1)
		outl p2,a		; (2)
		dec r0			; (1) point to low byte of SP
		mov a,@r0		; (1) 
		mov r1,a		; (1)
		add a,#254		; (2) carry generated here
		mov @r0,a		; (1)
		movx a,@r1		; (2)
		mov r4,a		; (1)
		inc r1			; (1)
		movx a,@r1		; (2)
		mov r5,a		; (1)
		jc dpop2_done		; (2) used here
		inc r0			; (1)
		mov a,@r0		; (1)
		dec a			; (1)
		mov @r0,a		; (1)
dpop2_done	

; FALLTHROUGH
;}}}
dpop ;{{{ pop cell from data stack and store in r2 r3 (22/26)
		mov r0,#SP+1	; (2)
		mov a,@r0	; (1)
		outl p2,a	; (2)
		dec r0		; (1) r0=&SP low byte
		mov a,@r0	; (1) r0=*r0
		mov r1,a	; (1)
		add a,#254	; (2) carry generated here
		mov @r0,a	; (1)
		movx a,@r1	; (2)
		mov r2,a	; (1)
		inc r1		; (1)
		movx a,@r1	; (2)
		mov r3,a	; (1)
		jc dpop_done	; (2) is used here
		inc r0		; (1)
		mov a,@r0	; (1)
		dec a		; (1)
		mov @r0,a	; (1)
dpop_done	ret		; (2)
;}}}
dstore ;{{{ store top of data stack from in r2 r3 to addr r6 r7 (14)
		mov a,r7	; (1)
		outl p2,a	; (2)
		mov a,r6	; (1)
		mov r0,a	; (1)
		mov a,r2	; (1)
		movx @r0,a	; (2)
		inc r0		; (1)
		mov a,r3	; (1)
		movx @r0,a	; (2)
		ret		; (2)
;}}}
interrupt ;{{{
		retr
;}}}
puthex ;{{{ print character in a as two-character hex number, useful for debugging
; output byte as hex
; destroys r0,r1,r2
		mov r2,a
		anl a,#$f0
		swap a
		add a,#246
		jnc digith
		add a,#7	; 'A' - '0'
digith		add a,#58	; '0' + 10
		call putchar
		mov a,r2
		anl a,#$0f
		add a,#246
		jnc digitl
		add a,#7	; 'A' - '0'
digitl		add a,#58	; '0' + 10
		call putchar
		ret
;}}}
saveIP ;{{{ save r4 r5 to IP
		mov r0,#IP	; (2) store IP
		mov a,r4	; (1)
		mov @r0,a	; (1)
		inc r0		; (1)
		mov a,r5	; (1)
		mov @r0,a	; (1)
		ret
;}}}
loadIP ;{{{ load r4 r5 from IP
		mov r0,#IP		; (2)
		mov a,@r0		; (1)
		mov r4,a		; (1)
		inc r0			; (1)
		mov a,@r0		; (1)
		mov r5,a		; (1)
		ret
;}}}

; Wordlists ;{{{
;
; Traditionally, Forth words form a backward chained list, called the
; dictionary or vocabulary.  The global pointer LATEST points to the
; beginning of the last word, the head of the chain, used both to change
; the status of this word and as value for the predecessor of a new word.
; The details of the data structure vary among different implementations.
;
; ANS Forth only requires that newer words must be found before older ones,
; allowing to redefine words, but hiding the details of the data structure.
; ANS Forth further calls it a wordlist.
;
; This implementation uses the following structure for words:
;
;   +-----------------+---------+--------- - - +----------------- - - +
;   | Link field (LF) | Length/ | Name         | Code field (CF)/Body |
;   |                 | Flags   |              | (aligned)            |
;   +---- 2 bytes ----+- 1 byte +- n bytes - - +- m bytes ------- - - +
;            ^
;            |
;            |
;   +-----------------+---------+--------- - - +----------------- - - +
;   | Link field (LF) | Length/ | Name         | Code field (CF)/Body |
;   |                 | Flags   |              | (aligned)            |
;   +---- 2 bytes ----+- 1 byte +- n bytes - - +- m bytes ------- - - +
;            ^
;            |_______________________
;                                    |
;   +-----------------------------+  |
;   | Link field address (LFA)  0 |--
;   +-----------------------------+
;   |                             |
;                ...
;   |                             |
;   +-----------------------------+
;   | Link field address (LFA) 31 |
;   +-----------------------------+
;
; The code field is aligned to a cell address.  Although not required by
; the MCS-48, this allows to access both bytes of a cell in less time.
; A pointer to it is called the code field address (CFA).
;
; The wordlist is hashed to speed up compilation.  Each of the 32 buckets
; contains the LFA (link field address) of the last word for this hash value.
; If given the name, the word WORDHASH computes the pointer to the bucket.
; As in traditional implementations, the variable LATEST contains the LFA
; of the last word defined for status modifications.  A pointer to the
; hash table is a wordlist identifier (WID).
;
; Forth words may be up to 31 characters long.  That leaves three
; bits in the length field used for flags:

;UNUSED_FLAG	equ $80
IMMEDIATE_FLAG	equ $40
HIDDEN_FLAG	equ $20
LENGTH_MASK	equ $1f

; Technically, there is no reason why code and names are stored together
; and statically, but this simple memory management fits well for small
; microprocessors.
;}}}

; Arithmetic, Single precision
;{{{ *
; ( n1 n2 -- product )

		if * AND 1
		reserve 1
		endi

word_MULT	dw 0
		db 1
		db '*'
MULT		call dpop2

; r6 r7 = r2 r3 * r4 r5
		mov r1,#16
		clr a
		mov r6,a
		mov r7,a

mulstep		mov a,r3		; shift r2 r3 right, LSB in carry
		rrc a
		mov r3,a
		mov a,r2
		rrc a
		mov r2,a
		jnc noadd
		mov a,r4		; r6 r7 += r4 r5
		add a,r6
		mov r6,a
		mov a,r5
		addc a,r7
		mov r7,a
noadd		mov a,r4		; r4 r5 <<= 1
		add a,r4
		mov r4,a
		mov a,r5
		addc a,r5
		mov r5,a
		djnz r1,mulstep

		mov a,r6
		mov r2,a
		mov a,r7
		mov r3,a
		call dpush
		call loadIP
		jmp next
;}}}

	reserve 26

;{{{ /MOD
; ( n1 n2 -- remainder quotient )

; Divide n1 by n2. n3 has the same sign as n2 or is zero.

		if * AND 1
		reserve 1
		endi

word_DIVMOD	dw word_MULT			; prev
		db 4			; len | flags
		db '/','M','O','D',0
DIVMOD		call dpop2

; r2 r3 = r2 r3 / r4 r5, r6 r7 = r2 r3 % r4 r5
		clr a
		mov r6,a
		mov r7,a
		mov r1,#16
		clr c
divloop		mov a,r2	; rotate r2 r3 left through carry
		rlc a
		mov r2,a
		mov a,r3
		rlc a
		mov r3,a
		mov a,r6	; rotate r6 r7 left through carry
		addc a,r6
		mov r6,a
		mov a,r7
		addc a,r7
		mov r7,a
		mov a,r6	; try to subtract r4 r5 from r6 r7 with carry
		cpl a
		addc a,r4
		cpl a
		mov r6,a
		mov a,r7
		cpl a
		addc a,r5
		cpl a
		mov r7,a
		jnc divcont	; subtraction was OK
		mov a,r6	; it was not, add r4 r5 to r6 r7 again (no carry)
		add a,r4
		mov r6,a
		mov a,r7
		addc a,r5
		mov r7,a
divcont		cpl c
		djnz r1,divloop
		mov a,r2	; rotate r2 r3 left through carry for last bit
		rlc a		; and exchange with r6 r7 at the same time
		xch a,r6
		mov r2,a
		mov a,r3
		rlc a
		xch a,r7
		mov r3,a

		call dpush
		mov a,r6
		mov r2,a
		mov a,r7
		mov r3,a
		call dpush
		call loadIP
		jmp next
;}}}
;{{{ - (77/81)
; ( n1 n2 -- n1-n2 )

		if * AND 1
		reserve 1
		endi

word_SUB	dw word_DIVMOD		; prev
		db 1			; len | flags
		db '-'
SUB		call dpopfetch		; (2+47/51)
		mov a,r2		; (1)
		cpl a			; (1)
		add a,r4		; (1)
		cpl a			; (1)
		mov r2,a		; (1)
		mov a,r3		; (1)
		cpl a			; (1)
		addc a,r5		; (1)
		cpl a			; (1)
		mov r3,a		; (1)
		call dstore		; (2+14)
		call loadIP
		jmp next		; (2)
;}}}
;{{{ +
; ( n1 n2 -- n1+n2 )

		if * AND 1
		reserve 1
		endi

word_ADD	dw word_SUB		; prev
		db 1			; len | flags
		db '+'
ADD		call dpopfetch
		mov a,r2
		add a,r4
		mov r2,a
		mov a,r3
		addc a,r5
		mov r3,a
		call dstore
		call loadIP
		jmp next
;}}}
;{{{ 1+
; ( n1 -- n1+1 )

		if * AND 1
		reserve 1
		endi

word_ADD1	dw word_ADD		; prev
		db 2			; len | flags
		db '1','+',0
ADD1		call dfetch
		inc r2
		mov a,r2
		jnz add1done
		inc r3
add1done	call dstore
		jmp next
;}}}
;{{{ 1-
; ( n1 -- n1-1 )

		if * AND 1
		reserve 1
		endi

word_SUB1	dw word_ADD1		; prev
		db 2			; len | flags
		db '1','-',0
SUB1		call dfetch
		mov a,r2
		jnz sub1low
		dec r3
sub1low		dec r2
		call dstore
		jmp next
;}}}
;{{{ (shift)
; ( n1 n2 -- n1 shifted by n2, n2 negative being right shift )

		if * AND 1
		reserve 1
		endi

word_PSHIFT	dw word_SUB1		; prev
		db 7			; len | flags
		db '(','S','H','I','F','T',')',0
PSHIFT		call dpopfetch
		mov a,r5
		jb7 rshift
		mov a,r4
		anl a,#$f
		mov r4,a
lshiftloop	clr c
		mov a,r2
		rlc a
		mov r2,a
		mov a,r3
		rlc a
		mov r3,a
		djnz r4,lshiftloop
		jmp shiftdone
rshift		mov a,r4
		cpl a
		inc a
		anl a,#$f
		mov r4,a
rshiftloop	clr c
		mov a,r3
		rrc a
		mov r3,a
		mov a,r2
		rrc a
		mov r2,a
		djnz r4,rshiftloop
shiftdone	call dstore
		call loadIP
		jmp next
;}}}

; Arithmetic, Numeric comparisons
;{{{ <= (including cmpret helper)
		if * AND 1
		reserve 1
		endi

word_SLE	dw word_PSHIFT		; prev
		db 2			; len | flags
		db '<','=',0
SLE		call dpopfetch
		call scmp16
		jz cmpret
		clr a
		jc cmpret
		cpl a

cmpret
; entry with a==0 being true and a==-1 being false (from Z signalling equality)
		cpl a
		mov r2,a
		mov r3,a
		call dstore
		call loadIP
		jmp next
;}}}
;{{{ <
		if * AND 1
		reserve 1
		endi

word_SLT	dw word_SLE		; prev
		db 1			; len | flags
		db '<'
SLT		call dpopfetch
		call scmp16
		clr a
		jc cmpret
		cpl a
		jmp cmpret
;}}}
;{{{ =
		if * AND 1
		reserve 1
		endi

word_NEQ	dw word_SLT		; prev
		db 1			; len | flags
		db '='
NEQ		call dpopfetch
		call scmp16
		jz cmpret
		mov a,#$ff
		jmp cmpret
;}}}
;{{{ U<=
		if * AND 1
		reserve 1
		endi

word_ULE	dw word_NEQ		; prev
		db 3			; len | flags
		db 'U','<','='
ULE		call dpopfetch
		call ucmp16
		jz cmpret
		clr a
		jc cmpret
		cpl a
		jmp cmpret
;}}}
;{{{ U<
		if * AND 1
		reserve 1
		endi

word_ULT	dw word_ULE		; prev
		db 2			; len | flags
		db 'U','<',0
ULT		call dpopfetch
		call ucmp16
		clr a
		jc ult_t
		cpl a
ult_t		jmp cmpret
;}}}

; Arithmetic, Bitwise operations
;{{{ AND
		if * AND 1
		reserve 1
		endi

word_AND	dw word_ULT		; prev
		db 3			; len | flags
		db 'A', 'N', 'D'
BINAND		call dpopfetch
		mov a,r2
		anl a,r4
		mov r2,a
		mov a,r3
		anl a,r5
		mov r3,a
		call dstore
		call loadIP
		jmp next
;}}}
;{{{ OR
		if * AND 1
		reserve 1
		endi

word_OR		dw word_AND		; prev
		db 2			; len | flags
		db 'O','R',0
BINOR		call dpopfetch
		mov a,r2
		orl a,r4
		mov r2,a
		mov a,r3
		orl a,r5
		mov r3,a
		call dstore
		call loadIP
		jmp next
;}}}
;{{{ XOR
		if * AND 1
		reserve 1
		endi

word_XOR	dw word_OR		; prev
		db 3			; len | flags
		db 'X','O','R'
BINXOR		call dpopfetch
		mov a,r2
		xrl a,r4
		mov r2,a
		mov a,r3
		xrl a,r5
		mov r3,a
		call dstore
		call loadIP
		jmp next
;}}}
;{{{ INVERT
		if * AND 1
		reserve 1
		endi

word_INVERT	dw word_XOR		; prev
		db 6			; len | flags
		db 'I','N','V','E','R','T',0
BININVERT	call dfetch
		mov a,r2
		cpl a
		mov r2,a
		mov a,r3
		cpl a
		mov r3,a
		call dstore
		call loadIP
		jmp next
;}}}

; Stack manipulation
;{{{ RP!
		if * AND 1
		reserve 1
		endi

word_RPSTORE	dw word_INVERT		; prev
		db 3			; len | flags
		db 'R','P','!'		; name
RPSTORE		call dpop
		mov r0,#RP
		mov a,r2
		mov @r0,a
		inc r0
		mov a,r3
		mov @r0,a
		jmp next
;}}}
;{{{ SP!
		if * AND 1
		reserve 1
		endi

word_SPSTORE	dw word_RPSTORE		; prev
		db 3			; len | flags
		db 'S','P','!'		; name
SPSTORE		call dpop
		mov r0,#SP
		mov a,r2
		mov @r0,a
		inc r0
		mov a,r3
		mov @r0,a
		jmp next
;}}}
;{{{ DUP
		if * AND 1
		reserve 1
		endi

word_DUP	dw word_SPSTORE		; prev
		db 3
		db 'D','U','P'		; name
DUP		call dfetch
		call dpush
		jmp next
;}}}
;{{{ ?DUP
		if * AND 1
		reserve 1
		endi

word_QDUP	dw word_DUP		; prev
		db 4
		db '?','D','U','P',0	; name
QDUP		call dfetch
		mov a,r2
		orl a,r3
		jz nodup
		call dpush
nodup		jmp next
;}}}
;{{{ DROP (10/14)
		if * AND 1
		reserve 1
		endi

word_DROP	dw word_QDUP		; prev
		db 4
		db 'D','R','O','P',0	; name
DROP		mov r0,#SP	; (2)
		mov a,@r0	; (1)
		add a,#254	; (2)
		mov @r0,a	; (1)
		jc drop_done	; (2)
		inc r0		; (1)
		mov a,@r0	; (1)
		dec a		; (1)
		mov @r0,a	; (1)
drop_done	jmp next	; (2)
;}}}
;{{{ SWAP
		if * AND 1
		reserve 1
		endi

word_SWAP	dw word_DROP		; prev
		db 4
		db 'S','W','A','P',0	; name
					; (48/49)
SWAP		mov r0,#SP+1		; (2) r2 r0 = SP
		mov a,@r0		; (1)
		mov r2,a		; (1)
		outl p2,a		; (2)
		dec r0			; (1)
		mov a,@r0		; (1)
		mov r0,a		; (1)

		movx a,@r0		; (2) load top of stack in r6 r7
		mov r6,a		; (1)
		inc r0			; (1) note r0 is incremented now
		movx a,@r0		; (2)
		mov r7,a		; (1)

		mov a,r0		; (1) a r1 = r2 r0 - 2
		add a,#253		; (2) - 3 really, because of r0 increment
		mov r1,a		; (1)
		jc swap_spdecdone	; (2)
		mov a,r2		; (1)
		dec a			; (1)
		outl p2,a		; (2)
swap_spdecdone
	
		movx a,@r1		; (2) exchange r6 r7 with cell below top
		xch a,r6		; (1)
		movx @r1,a		; (2)
		inc r1			; (1)
		movx a,@r1		; (2)
		xch a,r7		; (1)
		movx @r1,a		; (2)

		mov a,r2		; (1) write r6 r7 to top of stack
		outl p2,a		; (2)
		mov a,r7		; (1)
		movx @r0,a		; (2)
		dec r0			; (1)
		mov a,r6		; (1)
		movx @r0,a		; (2)

		jmp next		; (2)
;}}}
;{{{ OVER
		if * AND 1
		reserve 1
		endi

word_OVER	dw word_SWAP		; prev
		db 4
		db 'O','V','E','R',0	; name
OVER		call dpopfetch
		mov r0,#SP		; (2)
		inc @r0			; (1)
		inc @r0			; (1)
		mov a,@r0		; (1)
		jnz over_spinc_done	; (2)
		inc r0			; (1)
		inc @r0			; (1)
over_spinc_done	call dpush
		call loadIP
		jmp next
;}}}
;{{{ ROT
; ( n1 n2 n3 -- n2 n3 n1 )

		if * AND 1
		reserve 1
		endi

word_ROT	dw word_OVER		; prev
		db 3
		db 'R','O','T'		; name
ROT		call dpop2		; top: r4 r5 middle: r2 r3
		mov a,r2		; top: r4 r5 middle: r6 r7
		mov r6,a
		mov a,r3
		mov r7,a
		call dpop		; top: r4 r5 middle: r6 r7 bottom: r2 r3
		mov a,r6		; top: r4 r5 middle: r2 r3 bottom: r6 r7
		xch a,r2
		mov r6,a
		mov a,r7
		xch a,r3
		mov r7,a
		call dpush		; push middle
		mov a,r4
		mov r2,a
		mov a,r5
		mov r3,a
		call dpush		; push top
		mov a,r6
		mov r2,a
		mov a,r7
		mov r3,a
		call dpush		; push bottom
		call loadIP
		jmp next
;}}}
;{{{ >R
		if * AND 1
		reserve 1
		endi

word_TOR	dw word_ROT		; prev
		db 2
		db '>','R',0		; name
TOR		call dpop
		call rpush
		jmp next
;}}}
;{{{ R>
		if * AND 1
		reserve 1
		endi

word_FROMR	dw word_TOR		; prev
		db 2
		db 'R','>',0		; name
FROMR		call rpop
		call dpush
		jmp next
;}}}
;{{{ RDROP
		if * AND 1
		reserve 1
		endi

word_RDROP	dw word_FROMR		; prev
		db 5
		db 'R','D','R','O','P'	; name
RDROP		mov r0,#RP	; (2)
		inc @r0		; (1)
		inc @r0		; (1)
		mov a,@r0	; (1)
		jnz rdrop_inc_done ; (2)
		inc r0		; (1)
		inc @r0		; (1)
rdrop_inc_done	jmp next
;}}}
;{{{ 2RDROP
		if * AND 1
		reserve 1
		endi

word_TWORDROP	dw word_RDROP
		db 6
		db '2','R','D','R','O','P',0
TWORDROP	mov r0,#RP
		mov a,@r0
		add a,#4
		mov @r0,a
		jnc twordrop_done
		inc r0
		inc @r0
twordrop_done	jmp next
;}}}
;{{{ R@
		if * AND 1
		reserve 1
		endi

word_RFETCH	dw word_TWORDROP
		db 2
		db 'R','@',0
RFETCH		call fetchret
		jmp next
;}}}
;{{{ 2R@
; Copy cell pair x1 x2 from the return stack.
; ( -- x1 x2   R: x1 x2 -- x1 x2 )

		if * AND 1
		reserve 1
		endi

word_TWORFETCH	dw word_RFETCH
		db 3
		db '2','R','@'
TWORFETCH	call rpop
		mov a,r2
		mov r6,a
		mov a,r3
		mov r7,a
		call fetchret
		mov a,r6
		mov r2,a
		mov a,r7
		mov r3,a
		call rpush
		call dpush
		jmp next
;}}}

; Memory, Memory access
;{{{ @
		if * AND 1
		reserve 1
		endi

word_FETCH	dw word_TWORFETCH	; prev
		db 1			; len | flags
		db '@'
FETCH		call dfetch
		mov a,r3
		outl p2,a
		mov a,r2
		mov r0,a
		movx a,@r0
		mov r2,a
		inc r0
		movx a,@r0
		mov r3,a
		call dstore
		jmp next
;}}}
;{{{ C@
		if * AND 1
		reserve 1
		endi

word_CFETCH	dw word_FETCH		; prev
		db 2			; len | flags
		db 'C','@',0
CFETCH		call dfetch
		mov a,r3
		outl p2,a
		mov a,r2
		mov r0,a
		movx a,@r0
		mov r2,a
		clr a
		mov r3,a
		call dstore
 		jmp next
;}}}
;{{{ PC@
		if * AND 1
		reserve 1
		endi

word_PCFETCH	dw word_CFETCH		; prev
		db 3			; len | flags
		db 'P','C','@'
PCFETCH		call dfetch
		anl p1,#3       ; led_a16data=0 a17data=0 a17code=* multiplexer=*
		mov a,r3
		outl p2,a
		mov a,r2
		mov r0,a
		movx a,@r0
		mov r2,a
		clr a
		mov r3,a
	        anl p1,#7	; led_a16data=0 a17data=1 a17code=* multiplexer=*
	        orl p1,#4
		call dstore
 		jmp next
;}}}
;{{{ !
; ( value addr -- )

		if * AND 1
		reserve 1
		endi

word_STORE	dw word_PCFETCH		; prev
		db 1			; len | flags
		db '!'
STORE		call dpop2
		mov a,r5
		outl p2,a
		mov a,r4
		mov r0,a
		mov a,r2
		movx @r0,a
		inc r0
		mov a,r3
		movx @r0,a
		call loadIP
		jmp next
;}}}
;{{{ C!
; ( value addr -- )

		if * AND 1
		reserve 1
		endi

word_CSTORE	dw word_STORE		; prev
		db 2			; len | flags
		db 'C','!',0
CSTORE		call dpop2
		mov a,r5
		outl p2,a
		mov a,r4
		mov r0,a
		mov a,r2
		movx @r0,a
		call loadIP
		jmp next
;}}}
;{{{ PC!
; ( value addr -- )

		if * AND 1
		reserve 1
		endi

word_PCSTORE	dw word_CSTORE		; prev
		db 3			; len | flags
		db 'P','C','!'
PCSTORE		call dpop2
		anl p1,#3       ; a16data=0 a17data=0 a17code=* mux=*
		mov a,r5
		outl p2,a
		mov a,r4
		mov r0,a
		mov a,r2
		movx @r0,a
	        orl p1,#4	; a16data=0 a17data=1 a17code=* mux=*
		call loadIP
		jmp next
;}}}
; Memory, Address arithmetic
;{{{ CHARS
		if * AND 1
		reserve 1
		endi

word_CHARS	dw word_PCSTORE		; prev
		db 5			; len | flags
		db 'C','H','A','R','S'
CHARS		jmp next
;}}}
;{{{ CELLS
		if * AND 1
		reserve 1
		endi

word_CELLS	dw word_CHARS		; prev
		db 5			; len | flags
		db 'C','E','L','L','S'
CELLS		call dfetch
		mov a,r2
		clr c
		rlc a
		mov r2,a
		mov a,r3
		rlc a
		mov r3,a
		call dstore
		jmp next
;}}}
;{{{ ALIGNED
		if * AND 1
		reserve 1
		endi

word_ALIGNED	dw word_CELLS		; prev
		db 7			; len | flags
		db 'A','L','I','G','N','E','D'
ALIGNED		call dfetch
		mov a,r2
		jb0 fix
		jmp next
fix		inc a
		mov r2,a
		jnz aligndone
		inc r3
aligndone	call dstore
		jmp next
;}}}
;{{{ CHARINC
		if * AND 1
		reserve 1
		endi

; ( addr1 -- addr2 )

; Return address of next character

word_CHARINC	dw word_ALIGNED
		db 5
		db 'C','H','A','R','+'
CHARINC		mov r0,#SP+1	; (2)
		mov a,@r0	; (1)
		outl p2,a	; (2)
		dec r0		; (1)
		mov a,@r0	; (1)
		mov r0,a	; (1)
		movx a,@r0	; (2)
		inc a		; (1)
		movx @r0,a	; (2)
		jnz charinc_ok	; (2)
		inc r0		; (1)
		movx a,@r0	; (2)
		inc a		; (1)
		movx @r0,a	; (2)
charinc_ok	jmp next	; (2)
;}}}
;{{{ CELLINC
		if * AND 1
		reserve 1
		endi

; ( addr1 -- addr2 )

; Return address of next cell

word_CELLINC	dw word_CHARINC
		db 5
		db 'C','E','L','L','+'
CELLINC		call dfetch
		inc r2
		inc r2
		mov a,r2
		jnz cellinc_ok
		inc r3
cellinc_ok	call dstore
		jmp next
;}}}
;{{{ COMPARE
		if * AND 1
		reserve 1
		endi

; ( s1 l1 s2 l2 -- res )

; Compare the two strings and return -1 if the first is lexicographically
; smaller, 0 if they are equal and 1 is the first one is greater than the
; second.

; Implementing this in assembler has a large impact on interactive
; performance, because it is used to search the dictionary.

word_COMPARE	dw word_CELLINC
		db 7
		db 'C','O','M','P','A','R','E'
COMPARE		call dpop	; r6 r7 = l2
		mov a,r2
		mov r6,a
		mov a,r3
		mov r7,a
		call dpop2	; r4 r5 = s2, r2 r3 = l1
		mov a,r3	; compare r3 and r7
		cpl a
		add a,r7
		cpl a
		jz cmplow
		jc comparelt
comparegt	call dpop
comparegt2	mov r2,#1
		mov r3,#0
		jmp comparedone
comparelt	call dpop
comparelt2	mov r2,#255
		mov r3,#255
		jmp comparedone
cmplow		mov a,r2	; r3 and r7 equal, compare r2 and r6
		cpl a
		add a,r6
		cpl a
		jz comparestr
		jc comparelt
		jmp comparegt
comparestr	call dpop	; r2 and r6 equal, compare strings
		mov a,r2	; r0 r3 = s1, r1 r5 = s2, r6 r7 = len
		mov r0,a
		mov a,r4
		mov r1,a
compareloop	mov a,r6	; len zero?
		orl a,r7
		jnz comparechars
		clr a		; yes, return 0
		mov r2,a
		mov r3,a
		jmp comparedone
comparechars	mov a,r5	; compare one character
		outl p2,a
		movx a,@r1
		mov r2,a
		mov a,r3
		outl p2,a
		movx a,@r0
		cpl a
		add a,r2
		cpl a
		jz comparecont
		jc comparelt2
		jmp comparegt2
comparecont	inc r0		; increment r0 r3
		mov a,r0
		jnz compareinc1
		inc r3
compareinc1	inc r1		; increment r1 r5
		mov a,r1
		jnz compareinc2
		inc r5
compareinc2	mov a,r6	; dec len
		dec r6
		jnz compareloop
		dec r7
		jmp compareloop

comparedone	call dpush
		call loadIP
		jmp next
;}}}
;{{{ COUNT
; ( addr1 -- addr2 count )

; Return address of first character and length of counted string

; Implementing this in assembler has a large impact on interactive
; performance, because it is used to search the dictionary.

word_COUNT	dw word_COMPARE
		db 5
		db 'C','O','U','N','T'
COUNT		call saveIP
		call dfetch
		mov a,r3
		outl p2,a
		mov a,r2
		mov r0,a
		inc a
		mov r2,a
		jnz countincdone
		inc r3
countincdone	movx a,@r0
		mov r4,a
		call dstore
		mov a,r4
		mov r2,a
		clr a
		mov r3,a
		call dpush		
		call loadIP
		jmp next
;}}}

	reserve 20

; Control structures
;{{{ BRANCH (14)
		if * AND 1
		reserve 1
		endi

word_BRANCH	dw word_COUNT		; prev
		db 6			; len | flags
		db 'B','R','A','N','C','H',0
BRANCH		mov a,r5		; (1) r4 r5 = r4 r5 @
		outl p2,a		; (2)
		mov a,r4		; (1)
		mov r0,a		; (1)
		movx a,@r0		; (2)
		mov r4,a		; (1)
		inc r0			; (1)
		movx a,@r0		; (2)
		mov r5,a		; (1)
		jmp next		; (2)
;}}}
;{{{ 0BRANCH (35/40 if not zero, 42/46 if zero)
		if * AND 1
		reserve 1
		endi

word_ZBRANCH	dw word_BRANCH		; prev
		db 7			; len | flags
		db '0','B','R','A','N','C','H'
ZBRANCH		call dpop	; (2+22/26)
		mov a,r2	; (1)
		orl a,r3	; (1)
		jz BRANCH	; (2+14)
		inc r4		; (1)
		inc r4		; (1)
		mov a,r4	; (1)
		jnz zbranchinc_done	; (2)
		inc r5		; (1)
zbranchinc_done	jmp next	; (2)
;}}}
;{{{ EXIT (29/31)
		if * AND 1
		reserve 1
		endi

word_EXIT	dw word_ZBRANCH		; prev
		db 4			; len | flags
		db 'E','X','I','T',0
EXIT		call rpop		; (2+21/23)
		mov a,r2		; (1)
		mov r4,a		; (1)
		mov a,r3		; (1)
		mov r5,a		; (1)
		jmp next		; (2)
;}}}
;{{{ EXECUTE
		if * AND 1
		reserve 1
		endi

; ( xt -- )

word_EXECUTE	dw word_EXIT		; prev
		db 7			; len | flags
		db 'E', 'X', 'E', 'C', 'U', 'T', 'E'
EXECUTE		call dpop
		jmp executeword
;}}}
;{{{ DODOES
; This is the core of CREATE and DOES>.  Used at the beginning of a word, it
; pushes the current PC (pointing to the next instruction already) plus a cell
; to the data stack and jumps to BRANCH.
; CREATE compiles code with the branch target pointing to EXIT, storing further
; data or code after the target.  DOES> changes the branch target.

		if * AND 1
		reserve 1
		endi

word_DODOES	dw word_EXECUTE	; prev
		db 6			; len | flags
		db 'D','O','D','O','E','S',0
DODOES		mov a,r4
		add a,#2
		mov r2,a
		mov a,r5
		addc a,#0
		mov r3,a
		call dpush
		jmp BRANCH
;}}}

; Other I/O, Displaying characters and strings
;{{{ EMIT
		if * AND 1
		reserve 1
		endi

word_EMIT	dw word_DODOES		; prev
		db 4			; len | flags
		db 'E','M','I','T',0
EMIT		call dpop
		mov a,r2
		call putchar
		jmp next
;}}}
; Other I/O, Single key input
;{{{ KEY
		if * AND 1
		reserve 1
		endi

word_KEY	dw word_EMIT		; prev
		db 3			; len | flags
		db 'K', 'E', 'Y'
KEY		call getchar
		mov r2,a
		mov r3,#0
		call dpush
		jmp next
;}}}

; Literals
;{{{ LIT (43/46)
		if * AND 1
		reserve 1
		endi

word_LIT	dw word_KEY		; prev
		db 3			; len | flags
		db 'L', 'I', 'T'
LIT		mov a,r5		; (1) IP high byte
		outl p2,a		; (2)
		mov a,r4		; (1)
		mov r1,a		; (1)
		movx a,@r1		; (2)
		mov r2,a		; (1)
		inc r1			; (1)
		movx a,@r1		; (2)
		mov r3,a		; (1)
		call dpush		; (2+22/24)
		inc r4			; (1)
		inc r4			; (1)
		mov a,r4		; (1)
		jnz lit_done		; (2)
		inc r5			; (1)
lit_done	jmp next		; (2)
;}}}
;{{{ 0 (29/31)
		if * AND 1
		reserve 1
		endi

word_ZERO	dw word_LIT		; prev
		db 1			; len | flags
		db '0'
ZERO		clr a			; (1)
		mov r2,a		; (1)
		mov r3,a		; (1)
		call dpush		; (2+22/24)
		jmp next		; (2)
;}}}

;{{{ TICKS
		if * AND 1
		reserve 1
		endi

word_TICKS	dw word_ZERO		; prev
		db 5			; len | flags
		db 'T','I','C','K','S'
TICKS		dis tcnti
		mov a,t
		mov r2,a
		mov r0,#CLOCKTICKS
		mov a,@r0
		mov r3,a
		en tcnti
		call dpush
		jmp next
;}}}
;{{{ IRAM
		if * AND 1
		reserve 1
		endi

word_IRAM	dw word_TICKS		; prev
		db 4			; len | flags
		db 'I','R','A','M',0
IRAM		mov r3,#0
		mov r0,#IRAMLEN
		mov a,@r0
		mov r2,a
		call dpush
		jmp next
;}}}

; Switching Program Memory Segment
;
; The MCS-48 can address 2 KB program memory directly, which means the
; program counter wraps after 7ff to 0.  To address more, you need bank
; switching, either internal or external.  Simply switching the external
; segment by loading P1 means the next instruction comes from the new bank.
; To execute programs stored in RAM starting at address 0, like programs
; from ROM do, you need to switch the code segment with an instruction at
; 7ff -- letting the PC wrap around to 0, so the next instruction is
; fetched from the beginning of the new segment.

codesegswitch	org $7ff
		outl p1,a

; Forth Variables

		if * AND 1
		reserve 1
		endi

; order of SP0, RP0 and IP0 must be same as in internal RAM
var_SP0		dw $4000
var_RP0		dw $8000
var_IP0		dw STARTUNHASHED

var_DP		dw forthend
var_BASE	dw 10
var_FORTHWORDLIST
		dw 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0	; 32 cells
		dw 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
var_LATEST	dw 0
var_CURRENT	dw var_FORTHWORDLIST
var_STATE	dw 0
var_NTIB	dw 0
var_IN		dw 0
var_TIB		reserve 80

; Bootstrapping
;
; At this point, all further functions are written in Forth, but
; compiled manually, because we have just the core, but no interpreter
; or compiler yet.


; Boolean flags
;{{{ FALSE
		if 0
: FALSE 0 EXIT ;
		endi

word_FALSE	dw word_IRAM		; prev
		db 5			; len | flags
		db 'F','A','L','S','E'
FALSE		dw ZERO, EXIT
;}}}
;{{{ TRUE
		if 0
: TRUE -1 EXIT ;
		endi

word_TRUE	dw word_FALSE		; prev
		db 4			; len | flags
		db 'T','R','U','E',0
TRUE		dw LIT, -1, EXIT
;}}}

; Arithmetic, Single precision
;{{{ +!
; ( n addr -- )

		if 0
: +! DUP @ ROT + SWAP ! ;
: +! DUP >R @ + <R ! ;
		endi

word_ADDSTORE	dw word_TRUE		; prev
		db 2			; len | flags
		db '+','!',0
ADDSTORE	dw DUP, FETCH		; n addr value
		dw ROT, ADD		; addr sum
		dw SWAP, STORE, EXIT
;}}}
;{{{ NEGATE
		if 0
: NEGATE 0 SWAP - ;
		endi

word_NEGATE	dw word_ADDSTORE	; prev
		db 6			; len | flags
		db 'N','E','G','A','T','E',0
NEGATE		dw ZERO
		dw SWAP
		dw SUB
		dw EXIT
;}}}
;{{{ /
		if 0
: / /MOD SWAP DROP ;
		endi

word_DIV	dw word_NEGATE		; prev
		db 1			; len | flags
		db '/'
DIV		dw DIVMOD, SWAP, DROP
		dw EXIT
;}}}

; Arithmetic, Numeric comparisons
;{{{ >
		if 0
: > <= 0= ;
		endi

word_SGT	dw word_DIV		; prev
		db 1			; len | flags
		db '>'			; name
SGT		dw SLE			; code...
		dw ZEQ
		dw EXIT
;}}}
;{{{ >=
		if 0
: >= < 0= ;
		endi
word_SGE	dw word_SGT		; prev
		db 2			; len | flags
		db '>', '=',0		; name
SGE		dw SLT			; code...
		dw ZEQ
		dw EXIT
;}}}
;{{{ 0=
		if 0
: 0= 0 = ;
		endi

word_ZEQ	dw word_SGE		; prev
		db 2			; len | flags
		db '0','=',0
ZEQ		dw ZERO, NEQ
		dw EXIT
;}}}
;{{{ 0>
		if 0
: 0> 0 > ;
		endi

word_ZSGT	dw word_ZEQ		; prev
		db 2			; len | flags
		db '0','>',0
ZSGT		dw ZERO, SGT
		dw EXIT
;}}}
;{{{ 0<
		if 0
: 0< 0 < ;
		endi

word_ZSLT	dw word_ZSGT		; prev
		db 2			; len | flags
		db '0','<',0
ZSLT		dw ZERO, SLT
		dw EXIT
;}}}

; Stack manipulation, Data stack
;{{{ 2DUP
word_TWODUP	dw word_ZSLT		; prev
		db 4			; len | flags
		db '2','D','U','P',0	; name
TWODUP		dw OVER,OVER
		dw EXIT
;}}}
;{{{ NIP
word_NIP	dw word_TWODUP		; prev
		db 3			; len | flags
		db 'N','I','P'		; name
NIP		dw SWAP,DROP
		dw EXIT
;}}}
;{{{ TUCK
word_TUCK	dw word_NIP		; prev
		db 4			; len | flags
		db 'T','U','C','K',0	; name
TUCK		dw SWAP,OVER
		dw EXIT
;}}}

; Stack manipulation, Stack pointer manipulation
;{{{ RP0
word_RP0	dw word_TUCK		; prev
		db 3			; len | flags
		db 'R','P','0'		; name
RP0		dw LIT,var_RP0		; code...
		dw EXIT
;}}}
;{{{ SP0
word_SP0	dw word_RP0		; prev
		db 3			; len | flags
		db 'S','P','0'		; name
SP0		dw LIT,var_SP0		; code...
		dw EXIT
;}}}
;{{{ DP
word_DP		dw word_SP0		; prev
		db 2			; len | flags
		db 'D','P',0		; name
DP		dw LIT,var_DP		; code...
		dw EXIT
;}}}
;{{{ STATE
word_STATE	dw word_DP		; prev
		db 5			; len | flags
		db 'S','T','A','T','E'
STATE		dw LIT,var_STATE	; code...
		dw EXIT
;}}}
;{{{ [
word_LBRAC	dw word_STATE		; prev
		db 1 OR IMMEDIATE_FLAG	; len | flags
		db '['
LBRAC		dw FALSE,STATE,STORE
		dw EXIT
;}}}
;{{{ ]
word_RBRAC	dw word_LBRAC		; prev
		db 1			; len | flags
		db ']'
RBRAC		dw TRUE,STATE,STORE
		dw EXIT
;}}}

; Memory, Dictionary allocation
;{{{ HERE
; ( -- addr )
; returns dictionary pointer

word_HERE	dw word_RBRAC		; prev
		db 4			; len | flags
		db 'H','E','R', 'E', 0	; name
HERE		dw DP,FETCH
		dw EXIT
;}}}
;{{{ ALLOT
; ( n -- )
; Allocate n bytes in dictionary.

word_ALLOT	dw word_HERE		; prev
		db 5			; len | flags
		db 'A','L','L','O','T'
ALLOT		dw DP,ADDSTORE
		dw EXIT
;}}}
;{{{ COMMA
; ( cell -- )
; Append cell to dictionary.

word_COMMA	dw word_ALLOT		; prev
		db 1			; len | flags
		db ','
COMMA		dw HERE, STORE
		dw LIT, 1, CELLS, ALLOT
		dw EXIT
;}}}
;{{{ CCOMMA
; ( char -- )
; Append char to dictionary.

word_CCOMMA	dw word_COMMA		; prev
		db 2			; len | flags
		db 'C',',',0
CCOMMA		dw HERE, STORE
		dw LIT, 1, CHARS, ALLOT
		dw EXIT
;}}}
;{{{ LITERAL
; Compile code that returns a literal.

		if 0
: LITERAL IMMEDIATE ( n -- )
  LIT LIT , ,
;
		endi

word_LITERAL	dw word_CCOMMA
		db 7 OR IMMEDIATE_FLAG
		db 'L','I','T','E','R','A','L'
LITERAL		dw LIT, LIT, COMMA, COMMA
		dw EXIT
;}}}
;{{{ LATEST
word_LATEST	dw word_LITERAL		; prev
		db 6			; len | flags
		db 'L','A','T','E','S','T',0
LATEST		dw LIT,var_LATEST		; code...
		dw EXIT
;}}}
;{{{ SET-CURRENT
word_SETCURRENT	dw word_LATEST		; prev
		db 11			; len | flags
		db 'S','E','T','-','C','U','R','R','E','N','T'
SETCURRENT	dw LIT,var_CURRENT,STORE
		dw EXIT
;}}}
;{{{ GET-CURRENT
word_GETCURRENT	dw word_SETCURRENT	; prev
		db 11			; len | flags
		db 'G','E','T','-','C','U','R','R','E','N','T'
GETCURRENT	dw LIT,var_CURRENT,FETCH
		dw EXIT
;}}}

; Memory, Memory blocks 
;{{{ CMOVE
; ( src dst len -- )

; Move len bytes from src to dst, copying from lower to higher addresses.

word_CMOVE	dw word_GETCURRENT
		db 5
		db 'C','M','O','V','E'
CMOVE		dw DUP, ZBRANCH, cmovedone	; src dst len
		dw TOR, OVER, CFETCH		; src dst *src    R: len
		dw OVER, CSTORE			; src dst    R: len
		dw CHARINC, SWAP, CHARINC, SWAP	; src' dst'    R: len
		dw FROMR, SUB1			; src' dst' len'
		dw BRANCH, CMOVE		; src' dst' len'
cmovedone	dw DROP, DROP, DROP, EXIT
;}}}
;{{{ /STRING
; ( str1 len1 cutoff -- str2 len2 )

; Remove cutoff characters from the start of the string

word_TRIMSTRING	dw word_CMOVE
		db 7
		db '/','S','T','R','I','N','G'
TRIMSTRING	dw ROT, OVER			; len1 cutoff str1 cutoff
		dw CHARS, ADD			; len1 cutoff str2
		dw ROT, ROT			; str2 len1 cutoff
		dw SUB				; str2 len2
		dw EXIT
;}}}
;{{{ SKIP
; ( addr len match -- addr' len' )

; Skip matching characters, leave new address and len on the stack

word_SKIP	dw word_TRIMSTRING
		db 4
		db 'S','K','I','P',0
SKIP		dw TOR				; addr len    R: search
skiploop	dw DUP, ZBRANCH, skipend	; addr len    R: search
		dw SWAP, DUP, CFETCH		; len addr ch    R: search
		dw RFETCH, NEQ			; len addr match?    R: search
		dw ZBRANCH, skipdone		; len addr    R: match
		dw CHARINC, SWAP, SUB1		; addr len    R: match
		dw BRANCH, skiploop
skipdone	dw SWAP				; len addr
skipend		dw RDROP, EXIT			; addr len
;}}}
;{{{ SCAN
; ( addr len match -- addr' len' )

; Skip nonmatching characters, leave new address and len on the stack

word_SCAN	dw word_SKIP
		db 4
		db 'S','C','A','N',0
SCAN		dw TOR				; addr len    R: search
scanloop	dw DUP, ZBRANCH, scanend	; addr len    R: search
		dw SWAP, DUP, CFETCH		; len addr ch    R: search
		dw RFETCH, NEQ, ZEQ		; len addr cont?    R: search
		dw ZBRANCH, scandone		; len addr    R: match
		dw CHARINC, SWAP, SUB1		; addr len    R: match
		dw BRANCH, scanloop
scandone	dw SWAP				; addr len    R: match
scanend		dw RDROP, EXIT			; addr len
;}}}
;{{{ >COUNTED
; ( src n dst -- )

; copy to counted string

word_TOCOUNTED	dw word_SCAN
		db 8
		db '>','C','O','U','N','T','E','D',0
TOCOUNTED	dw TWODUP, CSTORE		; src n dst ( store length )
		dw CHARINC, SWAP		; src dst++ n
		dw CMOVE, EXIT			;
;}}}

; The input stream
;{{{ TIB
word_TIB	dw word_TOCOUNTED	; prev
		db 3			; len | flags
		db 'T','I','B'		; name
TIB		dw LIT,var_TIB		; code...
		dw EXIT
;}}}
;{{{ #TIB
word_NTIB	dw word_TIB		; prev
		db 4			; len | flags
		db '#','T','I','B',0	; name
NTIB		dw LIT,var_NTIB		; code...
		dw EXIT
;}}}
;{{{ >IN
word_IN		dw word_NTIB		; prev
		db 3			; len | flags
		db '>','I','N'		; name
IN		dw LIT,var_IN		; code...
		dw EXIT
;}}}
;{{{ WORD
; ( char -- addr )

; Look in the TIB for a word delimited by char, copy it to HERE as
; counted string, and append a not counted blank.  Return the address.
; If the delimiter is found at the end, >IN is incremented to skip it.

word_WORD	dw word_IN
		db 4
		db 'W','O','R','D',0
WORD		dw DUP,TIB,NTIB,FETCH,IN,FETCH	; char char adr len cur
		dw TRIMSTRING			; char char curadr currem ( delete up to current )
		dw DUP, TOR			; char char curadr currem    R: currem
		dw ROT, SKIP			; char wordstart startrem    R: currem
		dw OVER, TOR			; char wordstart startrem    R: currem wordstart
		dw ROT, SCAN			; wordend endrem    R: currem wordstart		
		dw FROMR, FROMR			; wordend endrem wordstart currem
		dw ROT, SUB			; wordend wordstart wordlen
		dw IN, FETCH, ADD		; wordend wordstart worddelim
		dw DUP, NTIB, FETCH, NEQ, ADD1, ADD	; wordend wordstart worddelimlen'
		dw IN, STORE			; wordend wordstart ( >IN points two chars after word, if possible )
		dw TUCK, SUB			; wordstart wordlen
		dw HERE, TOCOUNTED		; ( word copied as counted string )
		dw HERE, BL			; herebuf ' '
		dw HERE, COUNT, CHARS		; herebuf ' ' wordaddr wordlen
		dw ADD, CSTORE			; herebuf ( blank copied after word )
		dw EXIT
;}}}

; Word lists
;{{{ LFA>NAME
; ( lfa -- str len )

; Not a standard Forth word, it is used at various places to get from the
; LFA to the name.

word_LFA2NAME	dw word_WORD
		db 8
		db 'L','F','A','>','N','A','M','E',0
LFA2NAME	dw CELLINC		; nfa
		dw COUNT		; str len&flags
		dw LIT, $1f, BINAND	; str len
		dw EXIT
;}}}
;{{{ LFA>CFA
; ( lfa -- cfa )

; Not a standard Forth word, it is used at various places to get from the
; LFA to the CFA.

word_LFA2CFA	dw word_LFA2NAME
		db 7
		db 'L','F','A','>','C','F','A'
LFA2CFA		dw LFA2NAME
		dw ADD, ALIGNED		; cfa
		dw EXIT
;}}}
;{{{ WORDHASH
; ( str len wid -- addr )

; Return a pointer to the hash bucket for this word.

word_WORDHASH	dw word_LFA2CFA
		db 8
		db 'W','O','R','D','H','A','S','H',0
WORDHASH	dw TOR, TOR
		dw CFETCH, LIT, 223, BINAND 	; first char to upper case
		dw LIT, 2, PSHIFT		; shift left two times
		dw FROMR, BINXOR		; XOR with len
		dw DUP, LIT, -3, PSHIFT, BINXOR	; xor with self shifted right three times
		dw LIT,31,BINAND		; modulo hash size
		dw CELLS, FROMR, ADD		; get table offset
		dw EXIT
;}}}
;{{{ FORTH-WORDLIST
; ( -- wid )

word_FORTHWORDLIST	dw word_WORDHASH
		db 14
		db 'F','O','R','T','H','-','W','O','R','D','L','I','S','T',0
FORTHWORDLIST	dw LIT,var_FORTHWORDLIST,EXIT
;}}}
;{{{ SEARCH-WORDLIST
; ( str len wid -- 0 / xt 1 / xt -1 )

; If the counted string word is not found, result is word and the flag is 0.
; If it is, the result is the code field address.  The flag is 1 for IMMEDIATE
; words and -1 for normal words.

word_SEARCHWORDLIST	dw word_FORTHWORDLIST
		db 15
		db 'S','E','A','R','C','H','-','W','O','R','D','L','I','S','T'
SEARCHWORDLIST	dw TOR, TWODUP		; str len str len    R: wid
		dw FROMR,WORDHASH,FETCH	; str len curlfa
search		dw DUP, ZBRANCH, fail	; str len curlfa
		dw TOR, TWODUP, RFETCH	; str len str len curlfa   R: curlfa
		dw CELLINC, COUNT	; str len str len wstr wlen|flags    R: curlfa
		dw LIT, $3f, BINAND	; str len str len wstr wlen|hiddenflag    R: curlfa
		dw COMPARE		; str len different?    R: curlfa
		dw ZBRANCH, found	; str len     R: curlfa
		dw FROMR, FETCH		; str len curlfa ( advance to previous lfa )
		dw BRANCH, search	; str len nextlfa

fail		dw DROP, DROP, DROP, FALSE, EXIT	; 0

found		dw DROP, DROP, FROMR	; curlfa
		dw DUP, LFA2CFA		; curlfa curcfa
		dw SWAP			; curcfa curlfa
		dw CELLINC, CFETCH	; curcfa len&flags
		dw LIT, IMMEDIATE_FLAG, BINAND	; curcfa immediatebit
		dw ZEQ, LIT, 1, BINOR	; curcfa -1|1
		dw EXIT
;}}}
;{{{ FIND
; ( word -- addr 0/1/-1 )

; If the counted string word is not found, result is word and the flag is 0.
; If it is, the result is the code field address.  The flag is 1 for IMMEDIATE
; words and -1 for normal words.

word_FIND	dw word_SEARCHWORDLIST
		db 4
		db 'F','I','N','D',0
FIND		dw DUP, COUNT, FORTHWORDLIST, SEARCHWORDLIST ; word 0 / xt 1 / xt -1
		dw DUP,ZBRANCH,findfail
		dw ROT,DROP			; FALLTHROUGH

findfail	dw EXIT
;}}}

; Other I/O, Displaying characters and strings
;{{{ BL
word_BL		dw word_FIND		; prev
		db 2			; len | flags
		db 'B','L',0
BL		dw LIT, ' ', EXIT
;}}}
;{{{ SPACE
word_SPACE	dw word_BL		; prev
		db 5			; len | flags
		db 'S','P','A','C','E'
SPACE		dw BL, EMIT, EXIT
;}}}
;{{{ TYPE
; ( addr len -- )

; Output len characters starting at addr, if len is >0

word_TYPE	dw word_SPACE			; prev
		db 4				; len | flags
		db 'T','Y','P','E',0
TYPE		dw DUP, ZSGT			; addr len positive?
		dw ZBRANCH,typedone		; addr len
		dw OVER, CFETCH, EMIT		; addr len
		dw SUB1, SWAP, CHARINC, SWAP	; len-- addr++
		dw BRANCH, TYPE			; addr len
typedone	dw DROP, DROP, EXIT		;
;}}}
;{{{ CR
word_CR		dw word_TYPE		; prev
		db 2			; len | flags
		db 'C','R',0
CR		dw LIT, 13, EMIT, LIT, 10, EMIT
		dw EXIT
;}}}
; Other I/O, Line input and conversion
;{{{ BASE
word_BASE	dw word_CR		; prev
		db 4			; len | flags
		db 'B','A','S','E', 0	; name
BASE		dw LIT,var_BASE		; code...
		dw EXIT
;}}}
;{{{ DIGIT?
; ( c -- n -1 | n 0 )
word_DIGITQ	dw word_BASE			; prev
		db 6				; len | flags
		db 'D', 'I', 'G', 'I', 'T', '?', 0
DIGITQ		dw DUP				; if (c>'9')
		dw LIT,'9'
		dw SLE
		dw ZBRANCH
		dw letter
		dw LIT,'0'			; decimal: subtracting '0' maps '0' to 0
		dw SUB
		dw BRANCH
		dw range
letter		dw LIT,$df			; letter: change to lower case
		dw BINAND
		dw LIT,55			; subtracting 'A'-10 maps 'A' to 10
		dw SUB
range		dw DUP				; check range
		dw BASE
		dw FETCH
		dw ULT
		dw EXIT
;}}}
;{{{ NUMBER?
; ( a u -- d -1 | a u 0 )

word_NUMBERQ	dw word_DIGITQ			; prev
		db 7				; length | flags
		db 'N','U','M','B','E','R','?'
NUMBERQ		dw DUP, ZEQ			; addr len emptystring?
		dw ZBRANCH, notempty		; addr len
		dw FALSE, EXIT			; addr len 0
notempty	dw TWODUP			; addr len addr len
		dw OVER, CFETCH			; addr len addr len firstchar
		dw LIT, '-', NEQ		; addr len addr len neg?
		dw DUP, TOR			; addr len addr len neg?   R: neg?
		dw ADD				; addr len addr len-    R: neg?
		dw SWAP,RFETCH,CHARS,SUB,SWAP	; addr len addr+ len-    R: neg?
		dw DUP, ZBRANCH, fail1		; addr len addr+ len-    R: neg?
		dw ZERO, TOR			; addr len addr+ len-    R: neg? 0
convert		dw OVER, CFETCH, DIGITQ		; addr len addr+ len- n err    R: neg? num
		dw ZBRANCH, fail2		; addr len addr+ len- n    R: neg? num
		dw FROMR			; addr len addr+ len- n num    R: neg?
		dw BASE, FETCH, MULT, ADD	; addr len addr+ len- num    R: neg?
		dw TOR				; addr len addr+ len-    R: neg? num
		dw SWAP, CHARINC, SWAP		; addr len addr++ len-    R: neg? num
		dw SUB1				; addr len addr+ len--    R: neg? num
		dw DUP, ZEQ, ZBRANCH, convert   ; addr len addr+ len--    R: neg? num
		dw DROP, DROP, DROP, DROP	;     R: neg? num
		dw FROMR, FROMR			; num neg?
		dw ZBRANCH, nonegate		; num
		dw NEGATE			; num
nonegate	dw LIT, -1, EXIT		; num -1

fail2		dw DROP				; addr len addr+ len-    R: neg? num
		dw RDROP			; addr len addr+ len-    R: neg?
fail1		dw RDROP			; addr len addr+ len-
		dw DROP				; addr len addr+
		dw DROP				; addr len
		dw FALSE			; addr len 0
		dw EXIT
;}}}
;{{{ ACCEPT
; ( tib size -- count )

; Start line editor on the given buffer.
; Editing ends with Return.  There is no EOL terminator, an empty string
; simply is a string of 0 characters.

word_ACCEPT	dw word_NUMBERQ
		db 6
		db 'A','C','C','E','P','T',0
ACCEPT		dw OVER, SWAP, CHARS, ADD, TOR	; tib    R: end
		dw DUP, TOR			; tib    R: end cur
type		dw KEY				; tib key    R: end cur
		dw DUP, LIT, 8, NEQ		; tib key bs?    R: end cur
		dw ZBRANCH, nobackspace		; tib key    R: end cur
		dw DROP, DUP, RFETCH, NEQ	; tib first?    R: end cur
		dw ZEQ, ZBRANCH, type		; tib    R: end cur
		dw LIT, 8, EMIT			; tib    R: end cur
		dw BL, EMIT			; tib    R: end cur
		dw LIT, 8, EMIT			; tib    R: end cur
		dw FROMR, SUB1, TOR		; tib    R: end cur
		dw BRANCH, type			; tib    R: end cur
nobackspace	dw DUP, LIT, 13, NEQ		; tib key cr?    R: end cur
		dw OVER, LIT, 10, NEQ		; tib key cr? lf?   R: end cur
		dw BINOR			; tib key cr_or_lf?   R: end cur
		dw ZBRANCH, nocr		; tib key    R: end cur
		dw DROP				; tib    R: end cur
		dw CR				; tib    R: end cur
		dw FROMR, SWAP, SUB		; count    R: end
		dw RDROP			; count
		dw EXIT				; count
nocr		dw TWORFETCH			; tib key end cur    R: end cur
		dw NEQ				; tib key full?    R: end cur
		dw ZBRANCH, append		; tib key    R: end cur
		dw DROP, BRANCH, type		; tib    R: end cur
append		dw DUP, EMIT			; tib key    R: end cur
		dw RFETCH, CSTORE		; tib    R: end cur
		dw FROMR, CHARINC, TOR		; tib    R: end cur
		dw BRANCH, type
;}}}
;{{{ >DIGIT
; ( n -- c )

; convert to 0..9A..Z

word_TODIGIT	dw word_ACCEPT			; prev
		db 6				; len | flags
		db '>','D', 'I', 'G', 'I', 'T', 0
TODIGIT		dw DUP, LIT, 9, SGT		; n letter?
		dw ZBRANCH, digit		; n
		dw LIT, 'A'-10, ADD, EXIT	; c
digit		dw LIT, '0', ADD, EXIT		; c
;}}}
;{{{ ADDHASH (hidden word)
; ( str len -- )

; Add name field address to hashed dictionary.

ADDHASH		dw GETCURRENT, WORDHASH, DUP, FETCH ; lastLFAptr lastLFA
		dw LATEST, FETCH		; lastLFAptr lastLFA newLFA
		dw STORE			; lastLFAptr
		dw LATEST, FETCH, SWAP		; newLFA lastLFAptr
		dw STORE			;
		dw EXIT
;}}}
;{{{ (CREATE)
; --

; Create a new word definition.  First, store the LFA of the new word.
; Then, allocate the LF, but don't set it yet.  Read a word and abort,
; should it be empty or too long.  Get the pointer to the end of the
; link chain for this word, store it in the LF and change the end of
; the chain to the new word.  Allocate room for the name and align the
; data space pointer.

word_PCREATE	dw word_TODIGIT
		db 8
		db '(','C','R','E','A','T','E',')',0
PCREATE		dw HERE, LATEST			; LFA newLFAptr
		dw STORE			;
		dw LIT, 0, COMMA		;     allocate empty LF
		dw BL, WORD			; NFA
		dw DUP, CFETCH, DUP		; NFA namelen namelen
		dw ZBRANCH, createfail		; NFA namelen
		dw ADD1, ALIGNED, ALLOT		; NFA
		dw COUNT, ADDHASH		;
		dw EXIT
createfail	dw DROP,DROP			;
		dw LATEST, FETCH		; newLFA
		dw DP, STORE			;
		dw LIT, '?', EMIT		;
		dw EXIT				;
;}}}
;{{{ (EXIT)
; --

; Create a new word definition.  First, store the LFA of the new word.
; Then, allocate the LF, but don't set it yet.  Read a word and abort,
; should it be empty or too long.  Get the pointer to the end of the
; link chain for this word, store it in the LF and change the end of
; the chain to the new word.  Allocate room for the name and align the
; data space pointer.

word_PEXIT	dw word_PCREATE
		db 6
		db '(','E','X','I','T',')',0
PEXIT		dw EXIT
;}}}
;{{{ HIDE

; --

; Hide latest definition.

word_HIDE	dw word_PEXIT
		db 4
		db 'H','I','D','E',0
HIDE		dw LATEST, FETCH		; lfa
		dw CELLINC, DUP			; nfa nfa
		dw CFETCH			; nfa len|flags
		dw LIT, HIDDEN_FLAG, BINOR	; nfa len|flags|HIDDEN
		dw SWAP, CSTORE			;
		dw EXIT 
;}}}
;{{{ REVEAL

; --

; Reveal latest definition.

word_REVEAL	dw word_HIDE
		db 6
		db 'R','E','V','E','A','L',0
REVEAL		dw LATEST, FETCH		; lfa
		dw CELLINC, DUP			; nfa nfa
		dw CFETCH			; nfa len|flags
		dw LIT, NOT HIDDEN_FLAG, BINAND	; nfa len|flags&~HIDDEN
		dw SWAP, CSTORE			;
		dw EXIT 
;}}}
;{{{ IMMEDIATE

; --

; Make last definition immediate.

word_IMMEDIATE	dw word_REVEAL
		db 9
		db 'I','M','M','E','D','I','A','T','E'
IMMEDIATE	dw LATEST, FETCH		; lfa
		dw CELLINC, DUP			; nfa nfa
		dw CFETCH			; nfa len|flags
		dw LIT, IMMEDIATE_FLAG, BINOR	; nfa len|flags&~HIDDEN
		dw SWAP, CSTORE			;
		dw EXIT 
;}}}
;{{{ :

; --

; Define a new word.

word_COLON	dw word_IMMEDIATE
		db 1
		db ':'
COLON		dw PCREATE
		dw HIDE, RBRAC
		dw EXIT 
;}}}
;{{{ ;

; --

; End a word definition

word_SEMICOLON	dw word_COLON
		db 1 OR IMMEDIATE_FLAG
		db ';'
SEMICOLON	dw REVEAL, LIT, EXIT, COMMA	;
		dw LBRAC			;
		dw EXIT 
;}}}

; Miscellaneous Words
;{{{ INTERPRET
word_INTERPRET	dw word_SEMICOLON
		db 9
		db 'I','N','T','E','R','P','R','E','T'
INTERPRET	
intword		dw BL, WORD, DUP, CFETCH	; cstr len
		dw ZBRANCH, intdone		; cstr
		dw FIND, QDUP			; addr false | xt -1/1 -1/1
		dw ZBRANCH, noword		; addr | xt -1/1
		dw ADD1				; xt 0|2
		dw STATE, FETCH, ZEQ		; xt 0|2 interpreting?
		dw BINOR			; xt interpreting?
		dw ZBRANCH, compword		; xt
		dw EXECUTE			;
		dw BRANCH, intword		;
compword	dw COMMA, BRANCH, intword	;
noword		dw COUNT, NUMBERQ		; d -1 | str len 0
		dw ZBRANCH, nonumber		; d | str len
		dw STATE, FETCH			; d state
		dw ZBRANCH, intword		; d
		dw LITERAL			;
		dw BRANCH, intword		;
nonumber	dw TYPE, LIT, '?', EMIT, CR	;
		dw BRANCH, intword
intdone		dw DROP, EXIT
;}}}
;{{{ BANNER
word_BANNER	dw word_INTERPRET
		db 6
		db 'B','A','N','N','E','R',0
BANNER		dw LIT, signon, COUNT, TYPE
		dw EXIT
signon		db 29,13,10,'M','C','S','-','4','8',13,10
		db 'R','O','M',' ','R','e','v','.',' ','2','0','1','1','0','1','3','1',13,10
;}}}
;{{{ STARTUNHASHED, STARTHASHED (hidden words)
STARTUNHASHED
		dw LIT,word_last		; lfa
initloop	dw DUP,FETCH			; lfa lf
		dw SWAP				; lf lfa
		dw DUP				; lf lfa lfa
		dw LATEST			; lf lfa lfa newLFAptr
		dw STORE			; lf lfa
		dw LFA2NAME			; lf str len
		dw ADDHASH			; lf
		dw DUP,ZEQ,ZBRANCH,initloop	; lf  ( previous lfa )
		dw LIT,STARTHASHED,LIT,var_IP0	; STARTHASHED var_IP0
		dw STORE
STARTHASHED	dw BANNER
		dw QUIT
;}}}
;{{{ QUIT
word_last
word_QUIT	dw word_BANNER
		db 4			; len | flags
		db 'Q','U','I','T',0	; name
QUIT		dw RP0, FETCH, RPSTORE
		dw ZERO, STATE, STORE
quitloop	dw TIB, LIT, 80
		dw ACCEPT, NTIB, STORE
		dw ZERO, IN, STORE
		dw INTERPRET
		dw STATE,FETCH,ZEQ
		dw ZBRANCH,quitloop
		dw LIT,ok
		dw COUNT,TYPE
		dw BRANCH,quitloop
ok		db 3,'o','k',' '
;}}}

; At this point, you can define all further words in Forth and compile
; them on the host.

forthend
