\ If this module is ported to a faster CPU, add a 10 us delay here.
\
: i2c-delay ;

\ The start condition begins with SDA and SCL being high (or going high one
\ after another) followed by a falling edge of SDA and a subsequent falling
\ edge of SCL:
\     __
\ SDA   \__
\      __
\ SCL X  \_
\
: i2c-start ( -- )
  sda1! i2c-delay scl1! i2c-delay sda0! i2c-delay scl0! i2c-delay
;

\ The stop condition begins with SDA being (or going) low and SCL being
\ low, followed by a rising edge of SCL and a subsequent rising edge of SDA:
\        _
\ SDA __/
\       __
\ SCL _/
\
: i2c-stop ( -- )
  sda0! i2c-delay scl1! i2c-delay sda1! i2c-delay
;

\ A bit is sent by putting it on SDA and sending an active high clock pulse:
\
\ SDA XXXX
\       _
\ SCL _/ \
\
: i2c-putbit ( bit -- )
  i2c-delay sda! i2c-delay scl1! i2c-delay scl0!
;

\ A bit is received by sampling SDA before the falling edge of the clock
\ pulse.  Note that SDA must be high to release the bus.  The slave may
\ stretch the clock by pulling SCL low until it is ready to send the bit,
\ so the master has to wait until it sees the clock being high:
\
\ SDA XXXX
\       _
\ SCL _/ \
\
: i2c-getbit ( -- bit )
  sda1! scl1! BEGIN scl? UNTIL i2c-delay sda? scl0!
;

\ A byte is a sequence of 9 bit slots: 8 for sending the data byte from MSB
\ to LSB and one to receive the ACK bit.  This is identical for reading and
\ writing:
\
\ SDA XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
\       _   _   _   _   _   _   _   _   _
\ SCL _/7\_/6\_/5\_/4\_/3\_/2\_/1\_/0\_/A\

: i2c-putbyte ( byte -- nack )
  8 BEGIN
    SWAP
    DUP 128 AND i2c-putbit
    1 LSHIFT
    SWAP
  1- ?DUP 0= UNTIL
  DROP
  i2c-getbit
;

: i2c-getbyte ( nack -- byte )
  sda1! i2c-delay
  0
  8 BEGIN
    SWAP
    1 LSHIFT
    i2c-getbit
    IF 1 OR THEN
    SWAP
  1- ?DUP 0= UNTIL
  SWAP
  i2c-putbit
  sda1!
;
