\ Many RTCs store only the last two digits of the year, but have
\ an extra counter for the day of the week.  There is a simple algorithm
\ that computes the day of the week from a day, month and year, and
\ this algorithm can be reversed to cover a small range of centuries.
\
\ This code works for years from 1901 to 2299.  It needs to be modified
\ before March 2300 to cover a new set of centuries.

\ Map century {19..22} to offset.
\
: century2offset ( century -- offset )
  LIT [ HERE 0 , ]
  SWAP 19 - CELLS + @
  EXIT
  [ HERE 0 , 6 , 4 , 2 , SWAP ! ]
;

\ Map offset {0..6} to century.
\
: offset2century ( offset -- century )
  LIT [ HERE 0 , ]
  SWAP CELLS + @
  EXIT
  [ HERE 19 , -1 , 22 , -1 , 21 , -1 , 20 , SWAP ! ]
;

\ Map month {1..12} to day offset.
\
: month2offset ( month -- offset )
  LIT [ HERE 0 , ]
  SWAP 1- CELLS + @
  EXIT
  [ HERE 1 , 4 , 3 , 6 , 1 , 4 , 6 , 2 , 5 , 0 , 3 , 5 , SWAP ! ]
;

: toyyyy ( day mon yy wday -- day mon yyyy )
  OVER >R >R                            \ day mon yy   R: yy wday
  OVER 2 <= IF 99 + 100 /MOD DROP THEN  \ day mon ry   R: yy wday
  DUP 4 / +                             \ day mon tmp   R: yy wday
  >R OVER R> +                          \ day mon tmp   R: yy wday
  OVER month2offset +                   \ day mon tmp   R: yy wday
  R> 700 +                              \ day mon tmp wday+700  R: yy
  SWAP - 7 /MOD DROP                    \ day mon centoff   R: yy
  offset2century                        \ day mon century   R: yy
  R@ 0= IF OVER 2 <= IF 1+ THEN THEN    \ day mon century   R: yy
  100 * R> +                            \ day mon yyyy
;

: fromyyyy ( day mon yyyy --- day mon yy wday )
  DUP >R                                \ day mon yyyy    R: yyyy
  OVER 2 <= IF 1- THEN                  \ day mon ccyy'    R: yyyy
  100 /MOD                              \ day mon yy' cc'    R: yyyy
  century2offset                        \ day mon yy' coff    R: yyyy
  SWAP                                  \ day mon coff yy'    R: yyyy
  DUP 4 / + +                           \ day mon tmp    R: yyyy
  >R OVER OVER                          \ day mon day mon  R: yyyy tmp
  month2offset                          \ day mon day moff   R: yyyy tmp
  R> + + 7 /MOD DROP                    \ day mon wday   R: yyyy
  R> 100 /MOD DROP                      \ day mon wday yy
  SWAP                                  \ day mon yy wday
;
